import importlib
from enrichment.query_agent import QueryAgent

class KnowledgeSource:
    def __init__(self, name: str, description: str, functions_file: str, override_reliability: str, override_credibility: str):
        self.name = name
        self.description = description
        self.functions_file = functions_file
        self.override_reliability = override_reliability
        self.override_credibility = override_credibility
        self.tools = []
        self.query_agent = None

    def query_tool(self):
        functions_module = importlib.import_module(f"{self.functions_file[:-3]}")
        for func_name in dir(functions_module):
            if func_name == "load_dotenv":
                continue
            if not func_name.startswith('_'):                
                func = getattr(functions_module, func_name)
                if callable(func):
                    self.tools.append(func)

        self.query_agent = QueryAgent(self.tools)

        def query_handler(query_text):
            return self.query_agent.query(query_text)

        query_handler.__name__ = self.name+"_query"
        query_handler.__doc__ = self.description + \
            "ARGS: query_text: str - The query text to query the knowledge source\n" + \
            "RETURNS: str - The response from the knowledge source\n" + \
            "Override reliability: " + self.override_reliability + "\n" + \
            "Override credibility: " + self.override_credibility
        return query_handler

    def __str__(self) -> str:
        return self.name




from ontology.knowledge_ontology import KnowledgeOntology
from utils.modelconfig import my_query_model
from agno.agent import Agent
from textwrap import dedent
from datetime import datetime
from colored import cprint




class QueryRewriteAgent:
    def __init__(self, ontology: KnowledgeOntology, mytools: list):
        self.ontology = ontology        
        self.agent = Agent(
            name="Query rewrite agent",
            role="Rewrite the query to use known entities",
            model=my_query_model,
            tools=mytools,
            instructions=dedent("""
                Rewrite the query to use known entities.
                """),
                show_tool_calls=True,
                markdown=True,
                debug_mode=False,
            )
    
    def rewrite_query(self, query: str, class_entity_pairs: dict):
        #print(class_entity_pairs)
        #cprint(query, 'yellow')
        prompt = dedent(f"""
            Rewrite the query to use known entities from the graph, for entities not in the graph keep the entity names as is.
            Here are the known entity pairs: {class_entity_pairs}
            Here is the query to rewrite: {query}
            Only return the rewritten query, no other text.
            """
        )

        result = self.agent.run(prompt)
        #cprint(result.content, 'green')
        return result.content


from ontology.knowledge_ontology import KnowledgeOntology
from utils.modelconfig import my_model
from agno.agent import Agent
from textwrap import dedent
from datetime import datetime

class UpdateAgent:
    def __init__(self, ontology: KnowledgeOntology, mytools: list):
        self.ontology = ontology
        self.update_agent = Agent(
            name="Knowledge Graph Update Agent",
            role="Update the knowledge graph",
            model=my_model,
            tools=mytools,
            instructions=dedent(f"""
                The user is providing you unstrucutred knowledge. Translate the knowledge into a structured format based on the ontology.
                Ontology:[{self.ontology}]
                Return the results in RDFS format.
                When you are done, add every entity and relationship to the graph using the tools available to you 
                Today is {datetime.now().strftime("%Y-%m-%d")}
            """),
            show_tool_calls=True,
            markdown=True,
            debug_mode=False,
            )

    def update(self, knowledge: str):
        return self.update_agent.run("Translate the following knowledge into a structured format based on the ontology, then add every entity and relationship to the graph using the tools available to you.\n\n " + knowledge)



class Property:
    """Represents a property of an entity or relationship in the ontology."""
    def __init__(self, name: str, prop_type: str, description: str, primary_key: bool = False):
        """
        Initializes a Property object.

        Args:
            name (str): The name of the property.
            prop_type (str): The data type of the property (e.g., 'string', 'float').
            description (str): A description of the property.
            primary_key (bool): True if this property is the primary key for its entity.
        """
        self.property_name = name
        self.type = prop_type
        self.description = description
        self.primary_key = primary_key

    def __str__(self):
        """Returns a string representation of the property."""
        pk_str = " - Primary Key" if self.primary_key else ""
        return f"{self.property_name} ({self.type}) - {self.description}{pk_str}"



from ontology.property import Property


class EntityClass:
    """Represents a class of entities (nodes) in the ontology."""
    def __init__(self, name: str, description: str):
        """
        Initializes an EntityClass object.

        Args:
            name (str): The name of the entity class (e.g., 'Company').
            description (str): A description of the entity class.
        """
        self.entity_class_name = name
        self.description = description
        self.properties = []
        self.primary_key_prop = None

    def add_property(self, property: "Property"):
        """
        Adds a property to the entity class.

        Args:
            property (Property): The property to add.
        """
        self.properties.append(property)
        if property.primary_key:
            self.primary_key_prop = property

    def __str__(self):
        """Returns a string representation of the entity class."""
        entity_str = ""
        entity_str += f"{self.entity_class_name} ({self.description})\n"
        entity_str += "      Properties:\n"
        for prop in self.properties:
            entity_str += f"      - {prop}\n"
        return entity_str

    def get_add_or_update_tool(self, add_or_update_entity_func):
        """
        Creates a tool function for adding or updating an entity of this class.

        Args:
            add_or_update_entity_func (function): The function to call to add/update the entity in the graph.

        Returns:
            function: A tool function that can be used by an agent.
        """
        primary_key_prop = next((prop for prop in self.properties if prop.primary_key), self.properties[0] if self.properties else None)
        if not primary_key_prop:
            return None

        def func(**kwargs):
            properties = kwargs.get('kwargs', kwargs)
            return add_or_update_entity_func(self.entity_class_name, primary_key_prop.property_name, properties)

        func.__name__ = "add_or_update_" + self.entity_class_name + "_information"
        func.__doc__ = f"Add or update a {self.entity_class_name} entity. Primary key: {primary_key_prop.property_name}"
        func.__parameters__ = self.get_tool_parameters_schema()
        return func

    def get_tool_parameters_schema(self):
        """
        Builds the JSON schema for the parameters of the add/update tool.

        Returns:
            dict: A dictionary representing the JSON schema.
        """
        schema = {
            "type": "object",
            "properties": {},
            "required": []
        }
        for prop in self.properties:
            prop_type = "string"
            if prop.type == "float":
                prop_type = "number"
            elif prop.type == "integer":
                prop_type = "integer"
            
            schema["properties"][prop.property_name] = {
                "type": prop_type,
                "description": prop.description
            }
            schema["required"].append(prop.property_name)
        return schema
    
    def get_get_all_entity_tool(self, get_all_entity_func):
        """
        Creates a tool function for getting all entities of this class.

        Args:
            get_all_entity_func (function): The function to call to get all entities from the graph.

        Returns:
            function: A tool function that can be used by an agent.
        """
        def func():
            return get_all_entity_func(self.entity_class_name)

        func.__name__ = "get_all_"+self.entity_class_name+"_entities"
        func.__doc__ = f"Get all {self.entity_class_name} entities."
        func.__parameters__ = {"type": "object", "properties": {}}
        return func

    def get_get_entity_properties_tool(self, get_entity_properties_func):
        """
        Creates a tool function for getting the properties of a specific entity.

        Args:
            get_entity_properties_func (function): The function to call to get entity properties.

        Returns:
            function: A tool function that can be used by an agent.
        """
        def func(**kwargs):
            properties = kwargs.get('kwargs', kwargs)
            param_name = f"{self.primary_key_prop.property_name}"
            primary_key_value = properties.get(param_name)
            return get_entity_properties_func(self.entity_class_name, self.primary_key_prop.property_name, primary_key_value)

        func.__name__ = "get_"+self.entity_class_name+"_properties"
        func.__doc__ = f"Get a {self.entity_class_name} properties."
        
        param_name = f"{self.entity_class_name}_{self.primary_key_prop.property_name}"
        func.__parameters__ = {
            "type": "object",
            "properties": {
                param_name: {
                    "type": "string",
                    "description": f"The {self.primary_key_prop.property_name} of the {self.entity_class_name}"
                }
            },
            "required": [param_name]
        }
        return func
